@@ -1,4 +1,4 @@
-NUM_VERSION = (2013, 4)
+NUM_VERSION = (2013, 5)
 VERSION = ".".join(str(nv) for nv in NUM_VERSION)
 __version__ = VERSION
 
@@ -84,7 +84,7 @@ def runscript(mainpyfile, args=None, pre_run="", steal_output=False):
             pre_run_edit = urwid.Edit("", pre_run)
 
             result = dbg.ui.call_with_ui(dbg.ui.dialog,
-                urwid.ListBox([urwid.Text(
+                urwid.ListBox(urwid.SimpleListWalker([urwid.Text(
                     "Your PuDB session has ended.\n\n%s"
                     "Would you like to quit PuDB or restart your program?\n"
                     "You may hit 'q' to quit."
@@ -92,7 +92,7 @@ def runscript(mainpyfile, args=None, pre_run="", steal_output=False):
                     urwid.Text("\n\nIf you decide to restart, this command "
                     "will be run prior to actually restarting:"),
                     urwid.AttrMap(pre_run_edit, "value")
-                    ]),
+                    ])),
                 [
                     ("Restart", "restart"),
                     ("Examine", "examine"),
